/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/utils/scheduler', [
    'io.ox/office/tk/config',
    'io.ox/office/tk/utils',
    'io.ox/office/tk/utils/class',
    'io.ox/office/tk/container/valueset'
], function (Config, Utils, Class, ValueSet) {

    'use strict';

    // the forced interval time between timer callbacks, in milliseconds
    var INTERVAL = Config.AUTOTEST ? 10 : 20;

    // the maximum duration before a forced interval will be added, in milliseconds
    var SLICE = Config.AUTOTEST ? 1000 : 200;

    // whether to create extended descriptor objects for logging in automated tests
    var AUTOTEST = Config.AUTOTEST || Config.UNITTEST;

    // whether to bypass the timer scheduler, and use the browser timers directly
    var DEBUG_DIRECT_TIMERS = !AUTOTEST && Config.getDebugUrlFlag('office:direct-timers');

    // whether to show the debug panel for pending timers and deferreds
    var SHOW_SCHEDULER_PANE = Config.getDebugUrlFlag('office:scheduler-pane');

    // private global functions ===============================================

    var logPendingObjects = SHOW_SCHEDULER_PANE ? (function () {

        // key for local storage to save the height of the panel
        var STORAGE_KEY = 'office:scheduler-pane:height';
        // root node of the debug panel
        var rootNode = $('<div id="io-ox-office-scheduler-pane" class="io-ox-office-main" style="display:none;">').appendTo('body');
        // output window for pending timers
        var timerNode = $('<div class="output-window">').appendTo(rootNode);
        // output window for pending deferred objects
        var deferredNode = $('<div class="output-window">').appendTo(rootNode);
        // vertical resizer node
        var resizerNode = $('<div class="resizer" tabindex="-1">').appendTo(rootNode);
        // current total height of the panel
        var height = 0;

        // sets a new height to the debug panel
        function setHeight(newHeight) {
            height = newHeight;
            localStorage.setItem(STORAGE_KEY, height);
            rootNode.css({ height: height });
            $('#io-ox-core').css({ bottom: height });
            $(window).trigger('resize');
        }

        // set the initial height of the debug panel
        setHeight(parseInt(localStorage.getItem(STORAGE_KEY), 10) || 200);

        // initialize CSS formatting, and resizer tracking
        require(['io.ox/office/tk/utils/tracking', 'less!io.ox/office/tk/utils/scheduler']).done(function (Tracking) {
            rootNode.css('display', '');
            Tracking.enableTracking(resizerNode);
            resizerNode.on('tracking:move', function (evt) {
                setHeight(Utils.minMax(30, Math.round(window.innerHeight / 2), height - evt.moveY));
            });
        });

        // fills the passed target node with logging data
        function fillNode(targetNode, pendingObjects, titleLabel) {

            var lines = pendingObjects.map(function (pendingObj) {
                var markup = '<tr>';
                markup += '<td>' + pendingObj.uid + '</td>';
                markup += '<td>' + Utils.escapeHTML(pendingObj.windowId) + '</td>';
                markup += '<td>' + Utils.escapeHTML(pendingObj.logMessage) + '</td>';
                markup += '<td>' + (pendingObj.background ? 'background' : '') + '</td>';
                markup += '</tr>';
                return { markup: markup, uid: pendingObj.uid };
            }).sort(function (line1, line2) {
                return Utils.compareIndexedStrings(line1.uid, line2.uid);
            });

            var markup = '<p>' + pendingObjects.length + ' pending ' + titleLabel + '</p>';
            markup += '<table class="debug-table"><tr><th>uid</th><th>windowId</th><th>logMessage</th><th>priority</th></tr>';
            lines.forEach(function (line) { markup += line.markup; });
            markup += '</table>';
            targetNode[0].innerHTML = markup;
        }

        // the implementation of the logger method
        function logPendingObjects() {
            fillNode(timerNode, Scheduler.getPendingTimers(), 'timers');
            fillNode(deferredNode, Scheduler.getPendingDeferreds(), 'deferred objects');
        }

        // return the throttled logger method from local scope
        return _.throttle(logPendingObjects, 200);

    }()) : _.noop;

    // class PendingObj =======================================================

    /**
     * Internal representation of a pending timer or promise. Provides a unique
     * identifier that can be used to identify the timer or promise.
     *
     * @constructor
     *
     * @property {String} uid
     *  The globally unique identifier of the pending timer or promise.
     *
     * @param {String} idPrefix
     *  A prefix used to create the unique identifier of this instance.
     *
     * @param {String|BaseObject} windowId
     *  The identifier of the root window of the context application, or an
     *  object with a method 'getWindowId' that returns such a window
     *  identifier.
     *
     * @param {String} logMessage
     *  A description for the pending timer or pending promise that will be
     *  inserted into the debug log used in automated test environments.
     *
     * @param {Object} [logOptions]
     *  Optional parameters for debug logging:
     *  - {Boolean} [logOptions.background=false]
     *      Whether the pending timer or pending promise represents a
     *      long-running background task with low priority. These background
     *      tasks will be ignored when waiting for pending tasks in automated
     *      test environments.
     */
    var PendingObj = Class.extendable(function (idPrefix, windowId, logMessage, logOptions) {

        this.uid = _.uniqueId(idPrefix);
        this.windowId = windowId.getWindowId ? windowId.getWindowId() : windowId;
        this.logMessage = logMessage;
        this.background = !!(logOptions && logOptions.background);

    }); // class PendingObj

    // class TimerObj =========================================================

    /**
     * A simple descriptor object for a scheduled timer.
     *
     * @constructor
     *
     * @extends PendingObj
     *
     * @property {Number} timestamp
     *  The timestamp for the target time when the timer should fire.
     *
     * @property {Function} callback
     *  The callback function to be invoked when the timer fires.
     *
     * @param {String|BaseObject} windowId
     *  The identifier of the root window of the context application requesting
     *  the timer, or an object with a method 'getWindowId' that returns such a
     *  window identifier. Used for debugging and logging of running timers in
     *  automated test environments.
     *
     * @param {String} logMessage
     *  A description for the pending timer that will be inserted into the
     *  debug log used in automated test environments.
     *
     * @param {Object} [logOptions]
     *  Optional parameters for debug logging:
     *  - {Boolean} [logOptions.background=false]
     *      Whether the pending timer represents a long-running background task
     *      with low priority. These background tasks will be ignored when
     *      waiting for pending tasks in automated test environments.
     */
    var TimerObj = PendingObj.extend(function (callback, delay, windowId, logMessage, logOptions) {

        PendingObj.call(this, 'timer', windowId, logMessage, logOptions);
        this.timestamp = _.now() + delay;
        this.callback = callback;

    }); // class TimerObj

    // class TimerQueue =======================================================

    /**
     * An instance of this class contains a queue of timer data objects that
     * will be stored sorted by their target timestamps.
     *
     * @constructor
     */
    function TimerQueue() {

        // The queue of all pending timers requested by the scheduler. The
        // queue is sorted ascending by target timestamps, the first element in
        // the array will be fired first.
        this.queue = [];

        // The identifiers of all pending timers, as flag set. When cancelling
        // a timer, its identifier will be removed from this set, but it will
        // not yet be removed from the queue, in order to prevent performing a
        // linear search for the queue entry. When processing the queue, all
        // its entries that are not contained anymore in this set will be
        // ignored.
        this.set = {};

    } // class TimerQueue

    // private methods --------------------------------------------------------

    /**
     * Removes all leading timers from the queue that have been cancelled in
     * the meantime, i.e. that have been removed from the flag set.
     */
    TimerQueue.prototype._validate = function () {
        var timerData = null;
        while ((timerData = this.queue[0]) && !(timerData.uid in this.set)) {
            this.queue.shift();
        }
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns whether this queue is empty.
     *
     * @returns {Boolean}
     *  Whether this queue is empty.
     */
    TimerQueue.prototype.empty = function () {
        return this.queue.length === 0;
    };

    /**
     * Returns an array with the descriptors of all active pending timers
     * contained in this queue.
     *
     * @returns {Array<TimerObj>}
     *  An array with the descriptors of all active pending timers.
     */
    TimerQueue.prototype.values = function () {
        return this.queue.filter(function (timer) {
            return timer.uid in this.set;
        }, this);
    };

    /**
     * Returns the first timer in this queue, but does not remove the timer
     * from this queue.
     *
     * @returns {Timer|Null}
     *  The first timer in this queue; or null if the queue is empty.
     */
    TimerQueue.prototype.peek = function () {
        return this.queue[0] || null;
    };

    /**
     * Removes the first timer from this queue.
     *
     * @returns {Timer|Null}
     *  The first timer in this queue; or null if the queue is empty.
     */
    TimerQueue.prototype.pull = function () {

        // pull the first timer from the queue (nothing more to do, if the queue is empty)
        var timer = this.queue.shift();
        if (!timer) { return null; }

        // remove the timer from the flag set
        delete this.set[timer.uid];

        // remove all following cancelled timers from the queue, to ensure that always
        // a valid pending timer is in front of the timer queue
        this._validate();

        return timer;
    };

    /**
     * Inserts a new timer into this queue.
     *
     * @param {Timer} timer
     *  The timer to be inserted into this queue.
     *
     * @returns {Boolean}
     *  Whether the timer has been inserted in front of the queue.
     */
    TimerQueue.prototype.insert = function (timer) {

        // find the insertion index (after other pending timers with the same timestamp)
        var index = Utils.findLastIndex(this.queue, function (currData) {
            return currData.timestamp <= timer.timestamp;
        }, { sorted: true }) + 1;

        // insert the new timer into the queue, and update the flag set
        this.queue.splice(index, 0, timer);
        this.set[timer.uid] = true;

        // update the debug panel
        logPendingObjects();

        // return whether the timer has been inserted in front of the queue
        return index === 0;
    };

    /**
     * Removes the specified timer from this queue.
     *
     * @param {String} timerId
     *  The identifier of the timer to be removed from this queue.
     *
     * @returns {Boolean}
     *  Whether the timer has been found and removed from the front of the
     *  queue.
     */
    TimerQueue.prototype.remove = function (timerId) {

        // detect whether the timer is located infront of the queue
        var leading = !this.empty() && (this.queue[0].uid === timerId);

        // remove the identifier from the flag set, but not from the queue (no linear search)
        delete this.set[timerId];

        // remove all leading cancelled timers from the queue, to ensure that always
        // a valid pending timer is in front of the timer queue
        if (leading) { this._validate(); }

        // update the debug panel
        logPendingObjects();

        // return whether the timer was located in front of the queue
        return leading;
    };

    // static methods ---------------------------------------------------------

    /**
     * Creates a system timer for the next pending timer in a timer queue that
     * will invoke the callback function associated to that timer.
     *
     * @param {TimerQueue} queue
     *  The timer queue whose first pending timer will be fired next.
     */
    TimerQueue.restartTimerLoop = (function () {

        // identifier of the system timer that will fire the next pending timer in the queue
        var sysTimerId = null;
        // the end timestamp of the current time slice where timer will be fired immediately
        var currSliceEnd = -1;
        // the end timestamp of the current time slice where timer will be fired immediately
        var lastTimerEnd = -1;
        // true while an external timer callback is running (needed to detect re-entrance)
        var firing = false;

        // implementation of the method restartTimerLoop() returned from local scope
        function restartTimerLoop(queue) {

            // do nothing if a timer is currently firing, i.e. a timer is being created from the
            // callback of another timer (this means that this function is currently running and
            // will create another system timer before leaving)
            if (firing) { return; }

            // cancel a running system timer (may have a wrong delay time, after inserting a short timer)
            if (sysTimerId) {
                window.clearTimeout(sysTimerId);
                sysTimerId = null;
            }

            // get the first pending timer from the queue, but do not remove it yet
            // (a faster timer may be added before the queued timer fires)
            var timer = queue.peek();
            if (!timer) { return; }

            // calculate the delay time for the next timer according to its desired timestamp
            var t0 = _.now();
            var delay = Math.max(0, timer.timestamp - t0);

            // if the timer wants to fire in the current synchronous time slice, and the current
            // timestamp is still inside the time slice, the timer will be fired without extra delay
            if ((currSliceEnd <= t0) || (currSliceEnd <= timer.timestamp)) {
                delay = Math.max(delay, lastTimerEnd + INTERVAL - t0);
            }

            // create the system timer that will fire the next pending timer from the queue
            sysTimerId = window.setTimeout(function () {

                // immediately reset the identifier to indicate that no system timer is running anymore
                sysTimerId = null;

                // extract the leading timer data from the queue, and update the flag set
                timer = queue.pull();

                // start a new synchronous time slice, if the last one has finished
                var t0 = _.now();
                if (currSliceEnd < t0) {
                    currSliceEnd = t0 + SLICE;
                }

                // update the debug panel
                logPendingObjects();

                // execute the callback function associated to the timer
                // bug 48203: also clean up after callback has thrown
                try {
                    firing = true;
                    timer.callback();
                } finally {
                    firing = false;
                    lastTimerEnd = _.now();

                    // create a new system timer for the next pending timer in the queue
                    restartTimerLoop(queue);
                }

            }, delay);
        }

        return restartTimerLoop;
    }());

    // class DeferredObj ======================================================

    /**
     * A simple descriptor object for a pending deferred object.
     *
     * @constructor
     *
     * @extends PendingObj
     *
     * @param {String|BaseObject} windowId
     *  The identifier of the root window of the context application requesting
     *  the deferred object, or an object with a method 'getWindowId' that
     *  returns such a window identifier. Used for debugging and logging of
     *  running timers in automated test environments.
     *
     * @param {String} logMessage
     *  A description for the pending promise that will be inserted into the
     *  debug log used in automated test environments.
     *
     * @param {Object} [logOptions]
     *  Optional parameters for debug logging:
     *  - {Boolean} [logOptions.background=false]
     *      Whether the pending promise represents a long-running background
     *      task with low priority. These background tasks will be ignored when
     *      waiting for pending tasks in automated test environments.
     */
    var DeferredObj = PendingObj.extend(function (windowId, logMessage, logOptions) {

        PendingObj.call(this, 'deferred', windowId, logMessage, logOptions);

    }); // class DeferredObj

    // singletons =============================================================

    /**
     * The queue of all pending timers requested by the scheduler. The queue is
     * sorted ascending by target timestamps, the first element in the array
     * will be fired first.
     *
     * @type TimerQueue
     */
    var pendingTimerQueue = new TimerQueue();

    /**
     * The map with all pending deferred objects requested by the scheduler.
     *
     * @type ValueSet<DeferredObj>
     */
    var pendingDeferredSet = new ValueSet('uid');

    // static class Scheduler =================================================

    /**
     * Synchronizes multiple timer requests globally. All timer requests will
     * be inserted into an internal queue. The scheduler will ensure delay
     * intervals between timer callbacks if necessary, in order to allow the
     * browser to process its event queue, rendering queue, and to process
     * network responses.
     */
    var Scheduler = {};

    // constants --------------------------------------------------------------

    /**
     * The forced interval time between timer callbacks, in milliseconds.
     *
     * @constant
     * @type Number
     */
    Scheduler.INTERVAL = INTERVAL;

    /**
     * The maximum duration before a forced interval will be added, in
     * milliseconds.
     *
     * @constant
     * @type Number
     */
    Scheduler.SLICE = SLICE;

    // static methods ---------------------------------------------------------

    /**
     * This method is the replacement for the method window.setTimeout().
     *
     * @param {Function} callback
     *  The callback function that will be invoked when the timer fires.
     *
     * @param {Number} delay
     *  The delay time, in milliseconds.
     *
     * @param {String|BaseObject} windowId
     *  The identifier of the root window of the context application requesting
     *  the timer, or an object with a method 'getWindowId' that returns such a
     *  window identifier. Used for debugging and logging of running timers in
     *  automated test environments.
     *
     * @param {String} logMessage
     *  A description for the pending timer that will be inserted into the
     *  debug log used in automated test environments.
     *
     * @param {Object} [logOptions]
     *  Optional parameters for debug logging:
     *  - {Boolean} [logOptions.background=false]
     *      Whether the pending timer represents a long-running background task
     *      with low priority. These background tasks will be ignored when
     *      waiting for pending tasks in automated test environments.
     *
     * @returns {String}
     *  A globally unique identifier for the timer. Intended to be used with
     *  the method Scheduler.clearTimeout().
     */
    Scheduler.setTimeout = DEBUG_DIRECT_TIMERS ? window.setTimeout.bind(window) : function (callback, delay, windowId, logMessage, logOptions) {

        // the data object with all settings of the new timer
        var timer = new TimerObj(callback, delay, windowId, logMessage, logOptions);
        // insert the new timer into the queue
        var leading = pendingTimerQueue.insert(timer);

        // if the new timer has been inserted in front of the queue, a system timer needs to
        // be started that will fire the new timer (either the queue was empty before, or a
        // longer-running system timer will be cancelled)
        if (leading) { TimerQueue.restartTimerLoop(pendingTimerQueue); }

        // return the identifier of the timer to the caller
        return timer.uid;
    };

    /**
     * This method is the replacement for the method window.clearTimeout().
     *
     * @param {String} timerId
     *  The identifier of a timer that has been created with the method
     *  Scheduler.setTimeout().
     */
    Scheduler.clearTimeout = DEBUG_DIRECT_TIMERS ? window.clearTimeout.bind(window) : function (timerId) {

        // remove the timer from the queue, detect whether it was the leading entry inthe queue
        var leading = pendingTimerQueue.remove(timerId);

        // create a new system timer that will fire the (new) first available timer request in the queue
        if (leading) { TimerQueue.restartTimerLoop(pendingTimerQueue); }
    };

    /**
     * Creates a new deferred object.
     *
     * @param {String|BaseObject} windowId
     *  The identifier of the root window of the context application requesting
     *  the new deferred object, or an object with a method 'getWindowId' that
     *  returns such a window identifier. Used for debugging and logging of
     *  pending deferred objects in automated test environments.
     *
     * @param {String} logMessage
     *  A description for the pending promise that will be inserted into the
     *  debug log used in automated test environments.
     *
     * @param {Object} [logOptions]
     *  Optional parameters for debug logging:
     *  - {Boolean} [logOptions.background=false]
     *      Whether the pending promise represents a long-running background
     *      task with low priority. These background tasks will be ignored when
     *      waiting for pending tasks in automated test environments.
     *
     * @returns {jQuery.Deferred}
     *  The new deferred object.
     */
    Scheduler.createDeferred = function (windowId, logMessage, logOptions) {

        // the data object with all settings of the new deferred object
        var pending = new DeferredObj(windowId, logMessage, logOptions);
        pendingDeferredSet.insert(pending);
        logPendingObjects();

        return new $.Deferred().always(function () {
            pendingDeferredSet.remove(pending);
            logPendingObjects();
        });
    };

    /**
     * Returns an array with the descriptors of all pending timers.
     *
     * @returns {Array<TimerObj>}
     *  An array with the descriptors of all pending timers.
     */
    Scheduler.getPendingTimers = function () {
        return pendingTimerQueue.values();
    };

    /**
     * Returns an array with the descriptors of all pending deferred objects.
     *
     * @returns {Array<DeferredObj>}
     *  An array with the descriptors of all pending deferred objects.
     */
    Scheduler.getPendingDeferreds = function () {
        return pendingDeferredSet.values();
    };

    // exports ================================================================

    return Scheduler;

});
