/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define([
    'io.ox/office/tk/container/valuemap'
], function (ValueMap) {

    'use strict';

    // class ValueMap ========================================================

    describe('Toolkit class ValueMap', function () {

        it('should exist', function () {
            expect(ValueMap).to.be.a('function');
        });

        // static methods -----------------------------------------------------

        describe('method "pluck"', function () {
            it('should exist', function () {
                expect(ValueMap).itself.to.respondTo('pluck');
            });
            it('should create a map from an array', function () {
                var map = ValueMap.pluck([{ p: 'a' }, { p: 'b' }, { p: 'c' }, { p: 'a', q: 42 }], 'p');
                expect(map).to.be.an.instanceof(ValueMap);
                expect(map._cont).to.deep.equal({ 0: 'a', 1: 'b', 2: 'c', 3: 'a' });
            });
            var OBJ = { a: { p: 1 }, b: { p: 2 }, c: { p: 3 }, d: { p: 4, q: 42 } };
            var EXP = { a: 1, b: 2, c: 3, d: 4 };
            it('should create a map from an object', function () {
                var map = ValueMap.pluck(OBJ, 'p');
                expect(map._cont).to.deep.equal(EXP);
            });
            it('should create a map from another map', function () {
                var map = ValueMap.pluck(new ValueMap(OBJ), 'p');
                expect(map._cont).to.deep.equal(EXP);
            });
        });

        describe('method "rekey"', function () {
            it('should exist', function () {
                expect(ValueMap).itself.to.respondTo('rekey');
            });
            it('should create a map with property values as keys', function () {
                var map = ValueMap.rekey([{ p: 'a' }, { p: 'b' }], 'p');
                expect(map._cont).to.deep.equal({ a: { p: 'a' }, b: { p: 'b' } });
            });
        });

        // constructor --------------------------------------------------------

        describe('constructor', function () {
            it('should create a map from an array', function () {
                var map = new ValueMap(['a', 'b', 'c', 'a']);
                expect(map._cont).to.deep.equal({ 0: 'a', 1: 'b', 2: 'c', 3: 'a' });
            });
            var OBJ = { a: 1, b: 2, c: 3, d: 4 };
            it('should create a map from an object', function () {
                var map = new ValueMap(OBJ);
                expect(map._cont).to.deep.equal(OBJ);
            });
            it('should create a map from another map', function () {
                var map = new ValueMap(new ValueMap(OBJ));
                expect(map._cont).to.deep.equal(OBJ);
            });
        });

        // public methods -----------------------------------------------------

        describe('method "empty"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('empty');
            });
            it('should return whether the map is empty', function () {
                var map = new ValueMap();
                expect(map.empty()).to.equal(true);
                map.insert('a', 1);
                expect(map.empty()).to.equal(false);
            });
        });

        describe('method "single"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('single');
            });
            it('should return whether the map contains a single element', function () {
                var map = new ValueMap();
                expect(map.single()).to.equal(false);
                map.insert('a', 1);
                expect(map.single()).to.equal(true);
                map.insert('b', 2);
                expect(map.single()).to.equal(false);
            });
        });

        describe('method "clear"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('clear');
            });
            it('should clear the map', function () {
                var map = new ValueMap({ a: 1 });
                expect(map.empty()).to.equal(false);
                expect(map.clear()).to.equal(map);
                expect(map.empty()).to.equal(true);
            });
        });

        describe('method "has"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('has');
            });
            it('should return whether the element exists', function () {
                var map = new ValueMap({ a: 1 });
                expect(map.has('a')).to.equal(true);
                expect(map.has('b')).to.equal(false);
            });
        });

        describe('method "get"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('get');
            });
            it('should return an element', function () {
                var map = new ValueMap({ a: 1 });
                expect(map.get('a')).to.equal(1);
                expect(map.get('b')).to.equal(undefined);
                expect(map.get('a', null)).to.equal(1);
                expect(map.get('b', null)).to.equal(null);
            });
        });

        describe('method "getAny"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('getAny');
            });
            it('should return an element', function () {
                var map = new ValueMap();
                expect(map.getAny()).to.equal(undefined);
                map.insert('a', 1);
                map.insert('b', 2);
                map.insert('c', 3);
                var v1 = map.getAny();
                expect(v1).to.be.a('number');
                var v2 = map.getAny();
                expect(v2).to.equal(v1);
            });
        });

        describe('method "clone"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('clone');
            });
            var map = new ValueMap();
            map.insert('a', Object);
            it('should create a shallow clone', function () {
                var clone = map.clone();
                expect(clone.get('a')).to.equal(Object);
                expect(clone.get('b')).to.equal(undefined);
            });
            it('should create a deep clone', function () {
                var spy = sinon.spy(_.constant(42));
                var context = {};
                var clone = map.clone(spy, context);
                expect(clone.get('a')).to.equal(42);
                expect(clone.get('b')).to.equal(undefined);
                sinon.assert.alwaysCalledOn(spy, context);
            });
        });

        describe('method "insert"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('insert');
            });
            it('should insert map elements', function () {
                var map = new ValueMap();
                expect(map.has('a')).to.equal(false);
                expect(map.insert('a', Object)).to.equal(Object);
                expect(map.has('a')).to.equal(true);
                expect(map.get('a')).to.equal(Object);
                expect(map.insert('a', Array)).to.equal(Array);
                expect(map.get('a')).to.equal(Array);
                expect(map.insert('b', 42)).to.equal(42);
                expect(map.get('a')).to.equal(Array);
                expect(map.get('b')).to.equal(42);
            });
        });

        describe('method "remove"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('remove');
            });
            it('should remove map elements', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                expect(map.has('a')).to.equal(true);
                expect(map.remove('a')).to.equal(42);
                expect(map.has('a')).to.equal(false);
                expect(map.remove('c')).to.equal(undefined);
                expect(map.remove('c', null)).to.equal(null);
            });
        });

        describe('method "move"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('move');
            });
            it('should move map elements', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                expect(map.has('a')).to.equal(true);
                expect(map.move('a', 'c')).to.equal(map);
                expect(map.has('a')).to.equal(false);
                expect(map.get('c')).to.equal(42);
                expect(map.move('c', 'b')).to.equal(map);
                expect(map.get('b')).to.equal(42);
            });
        });

        describe('method "merge"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('merge');
            });
            it('should merge the maps', function () {
                var map1 = new ValueMap({ a: 42, b: 'abc' });
                var map2 = new ValueMap({ b: false, c: null });
                expect(map1.has('c')).to.equal(false);
                expect(map1.merge(map2)).to.equal(map1);
                expect(map1.get('a')).to.equal(42);
                expect(map1.get('b')).to.equal(false);
                expect(map1.get('c')).to.equal(null);
            });
        });

        describe('method "getOrCreate"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('getOrCreate');
            });
            it('should return existing or create new elements', function () {
                var map = new ValueMap({ a: 42 });
                var spy = sinon.spy(_.constant('abc'));
                var context = {};
                expect(map.getOrCreate('a', spy, context)).to.equal(42);
                sinon.assert.callCount(spy, 0);
                expect(map.getOrCreate('b', spy, context)).to.equal('abc');
                sinon.assert.callCount(spy, 1);
                sinon.assert.alwaysCalledOn(spy, context);
                sinon.assert.calledWithExactly(spy.getCall(0), 'b');
            });
        });

        describe('method "getOrConstruct"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('getOrConstruct');
            });
            it('should return existing or create new elements', function () {
                var map = new ValueMap({ a: 42 });
                expect(map.getOrConstruct('a', Object)).to.equal(42);
                expect(map.getOrConstruct('b', Object)).to.deep.equal({});
                expect(map.getOrConstruct('c', RegExp, 'a')).to.deep.equal(/a/);
                expect(map.getOrConstruct('d', Array, 1, 2)).to.deep.equal([1, 2]);
                expect(map.getOrConstruct('e', Array, 1, 2, 3)).to.deep.equal([1, 2, 3]);
                expect(map.getOrConstruct('f', Array, 1, 2, 3, 4)).to.deep.equal([1, 2, 3, 4]);
                expect(map.getOrConstruct('g', Array, 1, 2, 3, 4, 5, 6, 7, 8)).to.deep.equal([1, 2, 3, 4, 5, 6, 7, 8]);
            });
        });

        describe('method "with"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('with');
            });
            it('should invoke callback for existing elements', function () {
                var map = new ValueMap({ a: 42 });
                var spy = sinon.spy(_.constant('abc'));
                var context = {};
                expect(map.with('b', spy, context)).to.equal(undefined);
                sinon.assert.callCount(spy, 0);
                expect(map.with('a', spy, context)).to.equal('abc');
                sinon.assert.callCount(spy, 1);
                sinon.assert.alwaysCalledOn(spy, context);
                sinon.assert.calledWithExactly(spy.getCall(0), 42);
            });
        });

        describe('method "keys"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('keys');
            });
            it('should return all keys', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                var keys = map.keys();
                expect(keys).to.be.an('array');
                expect(keys.sort()).to.deep.equal(['a', 'b']);
            });
            it('should return an empty array for an empty map', function () {
                expect(new ValueMap().keys()).to.deep.equal([]);
            });
        });

        describe('method "values"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('values');
            });
            it('should return all values', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                var values = map.values();
                expect(values).to.be.an('array').with.length(2);
                expect(values.indexOf(42)).to.be.at.least(0);
                expect(values.indexOf('abc')).to.be.at.least(0);
            });
            it('should return an empty array for an empty map', function () {
                expect(new ValueMap().values()).to.deep.equal([]);
            });
        });

        describe('method "toObject"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('toObject');
            });
            it('should return the internal map', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                expect(map.toObject()).to.equal(map._cont);
            });
        });

        describe('method "invert"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('invert');
            });
            it('should exist', function () {
                var map1 = new ValueMap({ a: 'c', b: 'd' });
                var map2 = map1.invert();
                expect(map2).to.not.equal(map1);
                expect(map1.get('a', null)).to.equal('c');
                expect(map1.get('b', null)).to.equal('d');
                expect(map1.get('c', null)).to.equal(null);
                expect(map1.get('d', null)).to.equal(null);
                expect(map2.get('a', null)).to.equal(null);
                expect(map2.get('b', null)).to.equal(null);
                expect(map2.get('c', null)).to.equal('a');
                expect(map2.get('d', null)).to.equal('b');
            });
        });

        describe('method "iterator"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('iterator');
            });
            it('should return a property iterator', function () {
                var map = new ValueMap({ a: 'c', b: 'd' });
                function matcher(result) { return !result.done && /^(ac|bd)$/.test(result.key + result.value); }
                var it = map.iterator();
                expect(it).to.respondTo('next');
                expect(it.next()).to.satisfy(matcher);
                expect(it.next()).to.satisfy(matcher);
                expect(it.next()).to.deep.equal({ done: true });
            });
        });

        describe('method "keyIterator"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('keyIterator');
            });
            it('should return a key iterator', function () {
                var map = new ValueMap({ a: 'c', b: 'd' });
                function matcher(result) { return !result.done && /^(ac|bd)$/.test(result.value + result.element); }
                var it = map.keyIterator();
                expect(it).to.respondTo('next');
                expect(it.next()).to.satisfy(matcher);
                expect(it.next()).to.satisfy(matcher);
                expect(it.next()).to.deep.equal({ done: true });
            });
        });

        describe('method "forEach"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('forEach');
            });
            it('should visit all elements', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                var spy = sinon.spy(), context = {};
                expect(map.forEach(spy, context)).to.equal(map);
                sinon.assert.callCount(spy, 2);
                sinon.assert.alwaysCalledOn(spy, context);
                sinon.assert.calledWithExactly(spy, 42, 'a', map);
                sinon.assert.calledWithExactly(spy, 'abc', 'b', map);
            });
        });

        describe('method "forEachKey"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('forEachKey');
            });
            it('should visit all keys and elements', function () {
                var map = new ValueMap({ a: 42, b: 'abc' });
                var spy = sinon.spy(), context = {};
                expect(map.forEachKey(spy, context)).to.equal(map);
                sinon.assert.callCount(spy, 2);
                sinon.assert.alwaysCalledOn(spy, context);
                sinon.assert.calledWithExactly(spy, 'a', 42, map);
                sinon.assert.calledWithExactly(spy, 'b', 'abc', map);
            });
        });

        describe('method "some"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('some');
            });
        });

        describe('method "every"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('every');
            });
        });

        describe('method "map"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('map');
            });
        });

        describe('method "filter"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('filter');
            });
        });

        describe('method "reject"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('reject');
            });
        });

        describe('method "pluck"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('pluck');
            });
        });

        describe('method "sortBy"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('sortBy');
            });
            it('should sort by callback value', function () {
                var map = new ValueMap({ a: -2, b: -1 });
                var spy = sinon.spy(Math.abs);
                expect(map.sortBy(spy, 42)).to.deep.equal([-1, -2]);
                expect(spy.alwaysCalledOn(42)).to.equal(true);
            });
            it('should sort by property', function () {
                var map = new ValueMap({ a: { p: 2 }, b: { p: 1 } });
                expect(map.sortBy('p')).to.deep.equal([{ p: 1 }, { p: 2 }]);
            });
        });

        describe('method "mapObject"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('mapObject');
            });
        });

        describe('method "pick"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('pick');
            });
        });

        describe('method "omit"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('omit');
            });
        });

        describe('method "find"', function () {
            it('should exist', function () {
                expect(ValueMap).to.respondTo('find');
            });
        });
    });

    // ========================================================================
});
